#include "toglimg.h"


#define	_BI_RGB			0
#define	_BI_RLE8		1
#define	_BI_RLE4		2
#define	_BI_BITFIELDS	3
/* #define _BI_JPEG ? */

#define _ReadUL(p, fp)	fread((void *)p, sizeof(unsigned long), 1, fp)
#define _ReadL(p, fp)	fread((void *)p, sizeof(long), 1, fp)
#define _ReadUS(p, fp)	fread((void *)p, sizeof(unsigned short), 1, fp)
#define _WriteUL(p, fp)	fwrite((void *)p, sizeof(unsigned long), 1, fp)
#define _WriteL(p, fp)	fwrite((void *)p, sizeof(long), 1, fp)
#define _WriteUS(p, fp)	fwrite((void *)p, sizeof(unsigned short), 1, fp)

/*
bfSize          : 691256 : 14 + 40 + 480 * 480 * 3 = 691254
bfReserved1     : 0
bfReserved2     : 0
bfOffBits       : 54

biSize          : 40
biWidth         : 480
biHeight        : 480
biPlanes        : 1		must
biBitCount      : 24
biCompression   : 0
biSizeImage     : 0
biXPelsPerMeter : 2834
biYPelsPerMeter : 2834
biClrUsed       : 0
biClrImportant  : 0
*/

GLboolean toglCheckExtensionBMP(char filename[])
{
	return toglCheckExtension(filename, ".bmp");
}

GLboolean toglCheckImageBMP(char file_name[])
{
	FILE	*fp;
    char	sig[3];


	sig[2] = '\0';

	if ((fp = fopen(file_name, "rb")) == NULL) {
		return GL_FALSE;
	}

	if (fread(sig, sizeof(char), 2, fp) != 2) {
		fclose(fp);
		return GL_FALSE;
    }
	
	fclose(fp);

#ifdef	_DEBUG	
	fprintf(stderr, "MAGIC   : %s\n", sig);
#endif	/* _DEBUG */

    if (strncmp(sig, "BM", 2)) {
		return GL_FALSE;
    }

	return GL_TRUE;
}

GLboolean toglReadImageBMP(char filename[], TOGLimage *img)
{
	FILE				*fp;
	long				ll;

	/* BITMAPFILEHEADER */
    unsigned short		bfType;
    unsigned long		bfSize;
    unsigned short		bfReserved1;
    unsigned short		bfReserved2;
    unsigned long		bfOffBits;

	/* BITMAPINFOHEADER */
    unsigned long		biSize;
    long				biWidth;
    long				biHeight;
    unsigned short		biPlanes;
    unsigned short		biBitCount;
    unsigned long		biCompression;
    unsigned long		biSizeImage;
    long				biXPelsPerMeter;
    long				biYPelsPerMeter;
    unsigned long		biClrUsed;
    unsigned long		biClrImportant;

	unsigned char		*p, *buf;
	register int		x, y;
	size_t				bufsize;

	int					j, k;
	int					pp;
	unsigned char		*palette = NULL;


	if (toglCheckImageBMP(filename) == GL_FALSE) {
		return GL_FALSE;
	}

	if ((fp = fopen(filename, "rb")) == NULL) {
		return GL_FALSE;
	}

	ll = _ReadUS(&bfType, fp);
	ll += _ReadUL(&bfSize, fp);
	ll += _ReadUS(&bfReserved1, fp);
	ll += _ReadUS(&bfReserved2, fp);
	ll += _ReadUL(&bfOffBits, fp);
#ifndef _LITTLE_ENDIAN
	toglChangeEndianusv(&bfType);
	toglChangeEndianuiv(&bfSize);
	toglChangeEndianusv(&bfReserved1);
	toglChangeEndianusv(&bfReserved2);
	toglChangeEndianuiv(&bfOffBits);
#endif

#ifdef	_DEBUG
	fprintf(stderr, "bfSize      : %d\n", bfSize);
	fprintf(stderr, "bfReserved1 : %d\n", bfReserved1);
	fprintf(stderr, "bfReserved2 : %d\n", bfReserved2);
	fprintf(stderr, "bfOffBits   : %d\n", bfOffBits);
	fprintf(stderr, "\n");
#endif	/* _DEBUG */

	if (ll != 5) {
		fclose(fp);
		return GL_FALSE;
	}

	/* 'BM' */
	if(bfType != 0x4D42){ 
		fclose(fp);
		return GL_FALSE;
	} 


	ll = _ReadUL(&biSize, fp);
	ll += _ReadL(&biWidth, fp);
	ll += _ReadL(&biHeight, fp);
	ll += _ReadUS(&biPlanes, fp);
	ll += _ReadUS(&biBitCount, fp);
	ll += _ReadUL(&biCompression, fp);
	ll += _ReadUL(&biSizeImage, fp);
	ll += _ReadL(&biXPelsPerMeter, fp);
	ll += _ReadL(&biYPelsPerMeter, fp);
	ll += _ReadUL(&biClrUsed, fp);
	ll += _ReadUL(&biClrImportant, fp);

#ifndef _LITTLE_ENDIAN
	toglChangeEndianuiv(&biSize);
	toglChangeEndianiv(&biWidth);
	toglChangeEndianiv(&biHeight);
	toglChangeEndianusv(&biPlanes);
	toglChangeEndianusv(&biBitCount);
	toglChangeEndianuiv(&biCompression);
	toglChangeEndianuiv(&biSizeImage);
	toglChangeEndianiv(&biXPelsPerMeter);
	toglChangeEndianiv(&biYPelsPerMeter);
	toglChangeEndianuiv(&biClrUsed);
	toglChangeEndianuiv(&biClrImportant);
#endif

#ifdef	_DEBUG
	fprintf(stderr, "biSize          : %d\n", biSize);
	fprintf(stderr, "biWidth         : %d\n", biWidth);
	fprintf(stderr, "biHeight        : %d\n", biHeight);
	fprintf(stderr, "biPlanes        : %d\n", biPlanes);
	fprintf(stderr, "biBitCount      : %d\n", biBitCount);
	fprintf(stderr, "biCompression   : %d\n", biCompression);
	fprintf(stderr, "biSizeImage     : %d\n", biSizeImage);
	fprintf(stderr, "biXPelsPerMeter : %d\n", biXPelsPerMeter);
	fprintf(stderr, "biYPelsPerMeter : %d\n", biYPelsPerMeter);
	fprintf(stderr, "biClrUsed       : %d\n", biClrUsed);
	fprintf(stderr, "biClrImportant  : %d\n", biClrImportant);
	fprintf(stderr, "\n");
#endif	/* _DEBUG */

	if (ll != 11) {
		fclose(fp);
		return GL_FALSE;
	}


	if (biHeight <= 0) {
		fclose(fp);
		return GL_FALSE;
	}

	img->m_width = biWidth;
	img->m_height = biHeight;
	img->m_components = 3;
	
	if (biBitCount != 32 && biBitCount != 24 && biBitCount != 8 && biBitCount != 4 && biBitCount != 1) {
		fclose(fp);
		return GL_FALSE;
	}

	
	if (biCompression != _BI_RGB) {
		/* Not Support _BI_RLE8, _BI_RLE4, _BITFIELDS, _BI_JPEG */
		fclose(fp);
		return GL_FALSE;
	}

	/* pbg */
	switch (biBitCount) {
	case 1:		/* mN */
		if ((palette = (unsigned char *)calloc(sizeof(unsigned char) * 4, 2)) == NULL) {
			fclose(fp);
			return GL_FALSE;
		}
		if ((ll = fread(palette, sizeof(unsigned char) * 4, 2, fp)) != 2) {
			free(palette);
			fclose(fp);
			return GL_FALSE;
		}
		break;
	case 4:		/* 16F */
		if ((palette = (unsigned char *)calloc(sizeof(unsigned char) * 4, 16)) == NULL) {
			fclose(fp);
			return GL_FALSE;
		}
		if ((ll = fread(palette, sizeof(unsigned char) * 4, 16, fp)) != 16) {
			free(palette);
			fclose(fp);
			return GL_FALSE;
		}
		break;
	case 8:		/* 256F */
		if ((palette = (unsigned char *)calloc(sizeof(unsigned char) * 4, 256)) == NULL) {
			fclose(fp);
			return GL_FALSE;
		}
		if ((ll = fread(palette, sizeof(unsigned char) * 4, 256, fp)) != 256) {
			free(palette);
			fclose(fp);
			return GL_FALSE;
		}
		break;
	default:
		break;
	}

	/* */
	if (fseek(fp, bfOffBits, SEEK_SET) != 0) {
		if (palette) {
			free(palette);
		}
		fclose(fp);
		return GL_FALSE;
	}

	if (!toglAllocImage(img)) {
		if (palette) {
			free(palette);
		}
		fclose(fp);
		return GL_FALSE;
	}

	bufsize = (img->m_width * biBitCount) / 8;
	if ((bufsize % 4) != 0) {
		bufsize = ((bufsize / 4) + 1) * 4;
	}

	if ((buf = (unsigned char *)calloc(sizeof(unsigned char), bufsize)) == NULL) {
		if (palette) {
			free(palette);
		}
		fclose(fp);
		return GL_FALSE;
	}

	switch (biBitCount) {
	case 1:
		for(y = 0; y < img->m_height; y++) {
			if (fread(buf, sizeof(unsigned char), bufsize, fp) != bufsize) {
				free(buf);
				if (palette) {
					free(palette);
				}
				fclose(fp);
				return GL_FALSE;
			}

			p = img->m_buffer + img->m_width * y * 4;
			x = 0;
			for(j = 0; j < (long)bufsize; j++) {
				for (k = 0; k < 8; k++) {
					if (x < img->m_width) {
						pp = (buf[j] >> (7 - k)) & 0x01;
						*(p++) = palette[pp * 4 + 2];	/* R */
						*(p++) = palette[pp * 4];		/* G */
						*(p++) = palette[pp * 4 + 1];	/* B */
						*(p++) = 255;
					}
					x++;
				}
			}
		}
		break;
	case 4:
		for(y = 0; y < img->m_height; y++) {
			if (fread(buf, sizeof(unsigned char), bufsize, fp) != bufsize) {
				free(buf);
				if (palette) {
					free(palette);
				}
				fclose(fp);
				return GL_FALSE;
			}

			p = img->m_buffer + img->m_width * y * 4;
			x = 0;
			for(j = 0; j < (long)bufsize; j++) {
				if (x < img->m_width) {
					pp = (buf[j] >> 4) & 0x0f;
					*(p++) = palette[pp * 4 + 2];	/* R */
					*(p++) = palette[pp * 4];		/* G */
					*(p++) = palette[pp * 4 + 1];	/* B */
					*(p++) = 255;
				}
				x++;
				if (x < img->m_width) {
					pp = buf[j] & 0x0f;
					*(p++) = palette[pp * 4 + 2];	/* R */
					*(p++) = palette[pp * 4];		/* G */
					*(p++) = palette[pp * 4 + 1];	/* B */
					*(p++) = 255;
				}
				x++;
			}
		}
		break;
	case 8:
		for(y = 0; y < img->m_height; y++) {
			if (fread(buf, sizeof(unsigned char), bufsize, fp) != bufsize) {
				free(buf);
				if (palette) {
					free(palette);
				}
				fclose(fp);
				return GL_FALSE;
			}

			p = img->m_buffer + img->m_width * y * 4;
			for(x = 0; x < img->m_width; x++) {
				*(p++) = palette[buf[x] * 4 + 2];	/* R */
				*(p++) = palette[buf[x] * 4];		/* G */
				*(p++) = palette[buf[x] * 4 + 1];	/* B */
				*(p++) = 255;
			}
		}
		break;
	case 24:
		for(y = 0; y < img->m_height; y++) {
			if (fread(buf, sizeof(unsigned char), bufsize, fp) != bufsize) {
				free(buf);
				if (palette) {
					free(palette);
				}
				fclose(fp);
				return GL_FALSE;
			}

			p = img->m_buffer + img->m_width * y * 4;
			for(x = 0; x < img->m_width; x++) {
				*(p++) = buf[x * 3 + 2];	/* R */
				*(p++) = buf[x * 3 + 1];	/* G */
				*(p++) = buf[x * 3 + 0];	/* B */
				*(p++) = 255;
			}
		}
		break;
	case 32:
		for(y = 0; y < img->m_height; y++) {
			if (fread(buf, sizeof(unsigned char), bufsize, fp) != bufsize) {
				free(buf);
				fclose(fp);
				return GL_FALSE;
			}

			p = img->m_buffer + img->m_width * y * 4;
			for(x = 0; x < img->m_width; x++) {
				*(p++) = buf[x * 4 + 2];	/* R */
				*(p++) = buf[x * 4 + 1];	/* G */
				*(p++) = buf[x * 4 + 0];	/* B */
				*(p++) = 255;
			}
		}
		break;
	default:
		break;
	}

	free(buf);

	if (palette) {
		free(palette);
	}

	fclose(fp);

	return GL_TRUE;
}

GLboolean toglWriteImageBMP(char filename[], TOGLimage *img)
{
	FILE				*fp;
	long				ll;

	/* BITMAPFILEHEADER */
    unsigned short		bfType;
    unsigned long		bfSize;
    unsigned short		bfReserved1;
    unsigned short		bfReserved2;
    unsigned long		bfOffBits;

	/* BITMAPINFOHEADER */
    unsigned long		biSize;
    long				biWidth;
    long				biHeight;
    unsigned short		biPlanes;
    unsigned short		biBitCount;
    unsigned long		biCompression;
    unsigned long		biSizeImage;
    long				biXPelsPerMeter;
    long				biYPelsPerMeter;
    unsigned long		biClrUsed;
    unsigned long		biClrImportant;
	unsigned char		*p, *buf;
	register int		x, y;
	size_t				bufsize;


	if ((fp = fopen(filename, "wb")) == NULL) {
		return GL_FALSE;
	}

	bufsize = img->m_width * 3;
	if (((img->m_width * 3) % 4) != 0) {
		bufsize += 4 - (img->m_width * 3) % 4;
	}

	/* BITMAPFILEHEADER */
    bfType = 0x4D42;	/* 'BM' */
    bfSize = bufsize * img->m_height + 14 + 40 + 2;
    bfReserved1 = 0;
    bfReserved2 = 0;
    bfOffBits = 14 + 40;

#ifndef _LITTLE_ENDIAN
	toglChangeEndianuiv(&bfType);
	toglChangeEndianuiv(&bfSize);
	toglChangeEndianusv(&bfReserved1);
	toglChangeEndianusv(&bfReserved2);
	toglChangeEndianuiv(&bfOffBits);
#endif

	ll = _WriteUS(&bfType, fp);
	ll += _WriteUL(&bfSize, fp);
	ll += _WriteUS(&bfReserved1, fp);
	ll += _WriteUS(&bfReserved2, fp);
	ll += _WriteUL(&bfOffBits, fp);

	if (ll != 5) {
		fclose(fp);
		return GL_FALSE;
	}

	/* */

	/* BITMAPINFOHEADER */
    biSize = 40;
    biWidth = img->m_width;
    biHeight = img->m_height;
    biPlanes = 1;
    biBitCount = 24;
    biCompression = _BI_RGB;
    biSizeImage = 0;
    biXPelsPerMeter = 2834;
    biYPelsPerMeter = 2834;
    biClrUsed = 0;
    biClrImportant = 0;

#ifndef _LITTLE_ENDIAN
	toglChangeEndianuiv(&biSize);
	toglChangeEndianiv(&biWidth);
	toglChangeEndianiv(&biHeight);
	toglChangeEndianusv(&biPlanes);
	toglChangeEndianusv(&biBitCount);
	toglChangeEndianuiv(&biCompression);
	toglChangeEndianuiv(&biSizeImage);
	toglChangeEndianiv(&biXPelsPerMeter);
	toglChangeEndianiv(&biYPelsPerMeter);
	toglChangeEndianuiv(&biClrUsed);
	toglChangeEndianuiv(&biClrImportant);
#endif

	ll = _WriteUL(&biSize, fp);
	ll += _WriteL(&biWidth, fp);
	ll += _WriteL(&biHeight, fp);
	ll += _WriteUS(&biPlanes, fp);
	ll += _WriteUS(&biBitCount, fp);
	ll += _WriteUL(&biCompression, fp);
	ll += _WriteUL(&biSizeImage, fp);
	ll += _WriteL(&biXPelsPerMeter, fp);
	ll += _WriteL(&biYPelsPerMeter, fp);
	ll += _WriteUL(&biClrUsed, fp);
	ll += _WriteUL(&biClrImportant, fp);

	if (ll != 11) {
		fclose(fp);
		return GL_FALSE;
	}

	/* */

	if ((buf = (unsigned char *)calloc(sizeof(unsigned char), bufsize)) == NULL) {
		fclose(fp);
		return GL_FALSE;
	}

	for(y = 0; y < img->m_height; y++) {
		p = img->m_buffer + img->m_width * y * 4;
		for(x = 0; x < img->m_width; x++) {
			buf[x * 3 + 2] = *(p++);	/* R */
			buf[x * 3 + 1] = *(p++);	/* G */
			buf[x * 3 + 0] = *(p++);	/* B */
			p++;
		}

		if (fwrite(buf, sizeof(unsigned char), bufsize, fp) != bufsize) {
			free(buf);
			fclose(fp);
			return GL_FALSE;
		}
	}

	buf[0] = '\0';
	buf[1] = '\0';
	if (fwrite(buf, sizeof(unsigned char), 2, fp) != 2) {
		free(buf);
		fclose(fp);
		return GL_FALSE;
	}
	
	free(buf);

	fclose(fp);

	return GL_TRUE;
}

